# -*- coding: utf-8 -*-
"""
orm sql schema globals module.
"""

SMALL_INTEGER_MAX = 32767
SMALL_INTEGER_MIN = -32768
INTEGER_MAX = 2147483647
INTEGER_MIN = -2147483648
BIG_INTEGER_MAX = 9223372036854775807
BIG_INTEGER_MIN = -9223372036854775808
